/*  cog.c  (C) 1997 Mark A. Livingston */
/* Any non-commercial use of this code is permitted so long as */
/* the above copyright line appears in the file.  Please contact */
/* the author regarding permission for commercial use. */
#include <stdlib.h>
#include <math.h>
#include <quat.h>
#include "config.h"
#include "quatext.h"
#include "data.h"
#include "cog.h"

/*
** Attempts to reduce the distance between the center of gravity of the n
** samples q with weights w and locations samp_loc.  cog is the current
** location on input, but the new location on output.  curPt is the ideal
** location, i.e. the current point.
*/
int
cogReduce( q_xyz_quat_type *q, double *w, q_vec_type *samp_loc, int n,
               q_vec_type cog, /* input and output */
	       q_vec_type curPt )
{
    int i, j, new_ct, *mask, changed, tmpI;
    double cog_dist, new_dist, tot_weight, new_weight, tmpD;
    q_vec_type wgt_sum, new_sum, sum_change, new_cog, tmpV;
    q_type tmpQ;

    /* build mask of in/out and find total weight */
    mask = (int *) malloc( n * sizeof( int ) );
    tot_weight = 0.0;
    for( i = 0; i < n; i++ ) {
      mask[i] = 1; /* in */
      tot_weight += w[i];
    }

    /* find the cog distance */
    cog_dist = q_vec_distance( curPt, cog );
    /* build a temp copy of the cog that we can update */
    q_vec_scale( wgt_sum, tot_weight, cog );

    /* bubble sort the input in decreasing order of weight */
    do {
      changed = 0;
      for( i = 0; i < n - 1; i++ )
	if( w[i] < w[i+1] ) {
	  /* swap the values */
	  q_vec_copy( tmpV, q[i].xyz );
	  q_vec_copy( q[i].xyz, q[i+1].xyz );
	  q_vec_copy( q[i+i].xyz, tmpV );
	  q_copy( tmpQ, q[i].quat );
	  q_copy( q[i].quat, q[i+1].quat );
	  q_copy( q[i+1].quat, tmpQ );
	  /* swap the weights */
	  tmpD = w[i];
	  w[i] = w[i+1];
	  w[i+1] = tmpD;
	  /* swap the locations */
	  q_vec_copy( tmpV, samp_loc[i] );
	  q_vec_copy( samp_loc[i], samp_loc[i+1] );
	  q_vec_copy( samp_loc[i+1], tmpV );
	  /* no need to swap the mask values - they're all 1 anyway */
	  changed = 1;
	} /* end if and for(i) */
    } while( changed == 1 );

#ifdef DEBUG
    printf( "In:\n" );
    for( i = 0; i < n; i++ ) {
      qx_file_xyz_quat_print( stdout, &q[i] );
      q_vec_print( samp_loc[i] );
    }
    printf( "COG in  = [ %lf, %lf, %lf ] < %lf >\n",
	    cog[0], cog[1], cog[2], cog_dist );    
#endif

    /* find a best set - I think this is a local, not global min */
    do {
      changed = 0;
      for( i = 0; i < n; i++ ) {
	/* compute effect on COG of this point */
	q_vec_scale( sum_change, w[i], samp_loc[i] );

	if( mask[i] == 1 ) { /* if it's in the set, take it out */
	  q_vec_subtract( new_sum, wgt_sum, sum_change );
	  new_weight = tot_weight - w[i];
	} else { /* if it's out of the set, put it in */
	  q_vec_add( new_sum, wgt_sum, sum_change );
	  new_weight = tot_weight + w[i];
	}
	q_vec_scale( new_cog, 1.0 / new_weight, new_sum );

	/* does the change improve things? */
	new_dist = q_vec_distance( curPt, new_cog );
	if( new_dist < cog_dist ) { /* make the change */
	  mask[i] = 1 - mask[i]; /* 1 -> 0 and 0 -> 1 */
	  cog_dist = new_dist;
	  q_vec_copy( cog, new_cog );
	  tot_weight = new_weight;
	  q_vec_copy( wgt_sum, new_sum );
	  changed = 1;
	} /* end if( better ) */
      } /* end for( each point i ) */
    } while( changed == 1 );

    /* order arrays so all points in the set come before all those out */
    for( i = 0, j = n - 1; i <= j; i++ ) {
      while( mask[j] == 0 )
	j--;
      if( mask[i] == 0 ) {
	/* swap the values */
	q_vec_copy( tmpV, q[i].xyz );
	q_vec_copy( q[i].xyz, q[j].xyz );
	q_vec_copy( q[j].xyz, tmpV );
	q_copy( tmpQ, q[i].quat );
	q_copy( q[i].quat, q[j].quat );
	q_copy( q[j].quat, tmpQ );
	/* swap the weights */
	tmpD = w[i];
	w[i] = w[j];
	w[j] = tmpD;
	/* swap the locations */
	q_vec_copy( tmpV, samp_loc[i] );
	q_vec_copy( samp_loc[i], samp_loc[j] );
	q_vec_copy( samp_loc[j], tmpV );
	/* exchange the mask values */
	tmpI = mask[i];
	mask[i] = mask[j];
	mask[j] = tmpI;
      } /* end if( mask[i] == 0 ) */
    } /* end for( i,j ) */

#ifdef DEBUG
    printf( "Out:\n" );
    for( j = 0; j < i; j++ ) {
      qx_file_xyz_quat_print( stdout, &q[j] );
    }
    printf( "COG out = [ %lf, %lf, %lf ] < %lf >\n",
	    cog[0], cog[1], cog[2], cog_dist );    
#endif
    return i;
}
