/*  data.c  (C) 1997 Mark A. Livingston */
/* Any non-commercial use of this code is permitted so long as */
/* the above copyright line appears in the file.  Please contact */
/* the author regarding permission for commercial use. */
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <quat.h>
#include "quatext.h"
#include "file.h"
#include "config.h"
#include "data.h"

/* bounding box, centroid of samples */
q_vec_type minSamp, maxSamp, ctrSamp;
double radSamp;

q_vec_type Origin = { 0.0, 0.0, 0.0 };

/* for keeping track of why points got rejected */
#define fobPOSN_JITTER    1
#define fobANG_JITTER     2
#define farPOSN_JITTER    4
#define farANG_JITTER     8

/*
** Performs a canned rejection test and adds to running totals of counts of
** whether a point was accepted or rejected, and why if it was rejected.
** Returns true if the point was accepted.
*/
int
RejectionTest( ConfigRec *cR, double d1, double d2, double a1, double a2,
	       int *tot )
{
    int code = 0;

    if( d1 > cR->posMaxFOB )
      code += fobPOSN_JITTER;
    if( d2 > cR->posMaxFaro )
      code += farPOSN_JITTER;
    if( a1 > cR->oriMaxFOB )
      code += fobANG_JITTER;
    if( a2 > cR->oriMaxFaro )
      code += farANG_JITTER;

    tot[ code ]++;

    return( code == 0 );
}

double
KernelValue( double sigma, double dist )
{
    return exp( -0.5 * SQR( dist ) / SQR( sigma ) )/* /(TWOPI*SQR(sigma)) */;
}

int
ReadDataFiles( ConfigRec *cR, q_xyz_quat_type *asceFsens,
	       q_xyz_quat_type *faroFrrodu )
{
    int i, j, code;
    int ptsLeft, ptCount, ptLastCt;
    /* samples */
    q_matrix_type fram2asceMAT, faro2framMAT, rrodc2srodMAT;
    /* Transformations */
    q_xyz_quat_type asceFfram, framFfaro, srodFrrodc, rrodcFsrod;
    q_xyz_quat_type framFasce, faroFfram, rroduFfaro, rrodcFrrodu, rroduFrrodc;
    q_xyz_quat_type rroduFsens, faroFsens, framFsens, rrodcFsens, srodFsens;
    q_xyz_quat_type faroFrrodc, faroFsrod, framFsrod, asceFsrod;
    /* variances of samples for rejection tests, count of rejections */
    double diagA, diagB, angleA, angleB, sampTime;
    int rejPts[16]; /*  sum_{i=0}^{i=4} ( 4 reasons choose i ) */
    FILE *fobPOSnoiseVdist, *fobORInoiseVdist; /* for scatter plots */
    FILE *faroPOSnoiseVdist, *faroORInoiseVdist;
    double distOrigin;

    /* Initialize */
    q_set_vec( minSamp,  100.0,  100.0,  100.0 );
    q_set_vec( maxSamp, -100.0, -100.0, -100.0 );
    for( i = 0; i < 16; i++ )
      rejPts[i] = 0;

    /* data files for noise scatter plots */
    fobPOSnoiseVdist = fopen( "FOBposNoiseVSdist.dat", "w" );
    fobORInoiseVdist = fopen( "FOBoriNoiseVSdist.dat", "w" );
    faroPOSnoiseVdist = fopen( "FAROposNoiseVSdist.dat", "w" );
    faroORInoiseVdist = fopen( "FAROoriNoiseVSdist.dat", "w" );

    for( ptLastCt = ptCount = 0, j = 0; j < cR->numDataFiles; j++ ) {
      /* Read in transformation data and build quaternions for trans loop */
      if( code = OpenFile( cR->transFile[ j ] ) ) {
        fprintf( stderr, "Unable to open file '%s'\n", cR->transFile[ j ] );
        exit( -1 );
      } else if( cR->verbose )
        printf( "Opened transformation file '%s'\n", cR->transFile[ j ] );

      /* Frame to Faro */
      NextMatrix( faro2framMAT );
      q_row_matrix_to_xyz_quat( &framFfaro, faro2framMAT );
      q_xyz_quat_invert( &faroFfram, &framFfaro );
      /* Ascension to Frame */
      NextMatrix( fram2asceMAT );
      q_row_matrix_to_xyz_quat( &asceFfram, fram2asceMAT );
      q_xyz_quat_invert( &framFasce, &asceFfram );
      /* Uncorrected Rod to Corrected Rod */
      NextSample( &rroduFrrodc );
      q_xyz_quat_invert( &rrodcFrrodu, &rroduFrrodc );
      /* Corrected Rod to SROD */
      NextMatrix( rrodc2srodMAT );
      q_row_matrix_to_xyz_quat( &srodFrrodc, rrodc2srodMAT );
      q_xyz_quat_invert( &rrodcFsrod, &srodFrrodc );

      CloseFile( );

      if( cR->nPts[j] > 0 ) {
#ifdef ASCII_FILES
        if( code = OpenFile( cR->sampFile[j] ) )
#else
        if( code = OpenBFile( cR->sampFile[j] ) ) {
#endif
          fprintf( stderr, "Unable to open file '%s'\n", cR->sampFile[j] );
          exit( -1 );
        } else if( cR->verbose )
          printf( "Opened file '%s'\n", cR->sampFile[j] );

	/* Read nPts points from the file, transform them, and store them */
	for( ptsLeft = cR->nPts[j]; ptsLeft > 0; ptsLeft-- ) {
	  /* Read the POINTS from the file; interpolation is hidden in here */
#ifdef ASCII_FILES
	  NextPoint( &asceFsens[ptCount], &diagA, &angleA, cR->nSampFOB[j] );
	  NextPoint( &faroFrrodu[ptCount], &diagB, &angleB, cR->nSampFARO[j] );
#else
	  NextBPointPair( &asceFsens[ptCount], &diagA, &angleA,cR->nSampFOB[j],
			  &faroFrrodu[ptCount],&diagB,&angleB,cR->nSampFARO[j],
			  &sampTime );
#endif

	  /* write out stuff for noise scatter plots */
	  distOrigin = q_vec_distance( asceFsens[ptCount].xyz, Origin );
	  if( fobPOSnoiseVdist != NULL )
	    fprintf( fobPOSnoiseVdist, "%lf %lf\n", distOrigin, diagA );
	  if( fobORInoiseVdist != NULL )
	    fprintf( fobORInoiseVdist, "%lf %lf\n", distOrigin, atan(angleA ));
	  if( faroPOSnoiseVdist != NULL )
	    fprintf( faroPOSnoiseVdist, "%lf %lf\n", distOrigin, diagB );
	  if( faroORInoiseVdist != NULL )
	    fprintf( faroORInoiseVdist, "%lf %lf\n", distOrigin, atan(angleB));

	  /* Check that the points are "good" data points */
	  if( distOrigin < 0.4572 || /* FOB undefined under 18 inches */
	      RejectionTest( cR, diagA, diagB, angleA, angleB, rejPts ) ) {
            /* Data point is OK */
	    if( cR->verbose ) {
	      fprintf( cR->outfp, "FOB XYZ  = ( %lf %lf %lf ) < %lf >\n",
                       asceFsens[ ptCount ].xyz[Q_X],
                       asceFsens[ ptCount ].xyz[Q_Y],
                       asceFsens[ ptCount ].xyz[Q_Z], diagA );
	      fprintf( cR->outfp,"FOB QUAT = ( %lf %lf %lf %lf ) < %lf >\n",
		       asceFsens[ ptCount ].quat[Q_X],
		       asceFsens[ ptCount ].quat[Q_Y],
		       asceFsens[ ptCount ].quat[Q_Z],
		       asceFsens[ ptCount ].quat[Q_W], angleA );
	      fprintf( cR->outfp, "FARO XYZ = ( %lf %lf %lf ) < %lf >\n",
                       faroFrrodu[ ptCount ].xyz[Q_X],
                       faroFrrodu[ ptCount ].xyz[Q_Y],
                       faroFrrodu[ ptCount ].xyz[Q_Z], diagB );
	      fprintf( cR->outfp,"FARO QUAT= ( %lf %lf %lf %lf ) < %lf >\n",
		       faroFrrodu[ ptCount ].quat[Q_X],
		       faroFrrodu[ ptCount ].quat[Q_Y],
		       faroFrrodu[ ptCount ].quat[Q_Z],
		       faroFrrodu[ ptCount ].quat[Q_W], angleB );
	    }

            /* Build the desired transformation by traversing the loop */
            q_xyz_quat_invert( &rroduFfaro, &faroFrrodu[ ptCount ] );
            q_xyz_quat_compose( &framFsens , &framFasce, &asceFsens[ptCount] );
	    q_xyz_quat_compose( &faroFsens , &faroFfram   , &framFsens    );
	    q_xyz_quat_compose( &rroduFsens, &rroduFfaro  , &faroFsens    );
	    q_xyz_quat_compose( &rrodcFsens, &rrodcFrrodu , &rroduFsens   );
	    q_xyz_quat_compose( &srodFsens , &srodFrrodc  , &rrodcFsens   );

#if 0
	    if( srodFsens.quat[Q_W] < 0.0 ) {
	      srodFsens.quat[Q_X] = -srodFsens.quat[Q_X];
	      srodFsens.quat[Q_Y] = -srodFsens.quat[Q_Y];
	      srodFsens.quat[Q_Z] = -srodFsens.quat[Q_Z];
	      srodFsens.quat[Q_W] = -srodFsens.quat[Q_W];
	    }
#endif
	    q_normalize( srodFsens.quat, srodFsens.quat );

	    /* Now fix the Faro position reading to be at SROD */
	    q_xyz_quat_compose( &faroFrrodc,&faroFrrodu[ptCount],&rroduFrrodc);
	    q_xyz_quat_compose( &faroFsrod , &faroFrrodc   , &rrodcFsrod  );
	    q_xyz_quat_compose( &framFsrod , &framFfaro    , &faroFsrod   );
	    q_xyz_quat_compose( &asceFsrod , &asceFfram    , &framFsrod   );

#if 0
/* INSANITY CHECK */
	    {
	      q_matrix_type srod2asce, sens2asce, sens2srod;
	      q_vec_type sens_srod_POS, sens_srod_ORI, diff;
	      q_xyz_quat_type asceFsrod_ORI, sensFsrod;
	      q_type diffQuat;

	      q_xyz_quat_to_row_matrix( srod2asce, &asceFsrod );
	      q_xyz_quat_to_row_matrix( sens2asce, &asceFsens[ptCount] );
	      q_xyz_quat_to_row_matrix( sens2srod, &srodFsens );

	      q_vec_scale( sens_srod_ORI, -1.0, sens2srod[3] );
	      qx_vec0_mult_matrix( sens_srod_ORI, sens_srod_ORI, srod2asce );
	      q_vec_subtract( sens_srod_POS, srod2asce[3], sens2asce[3] );

	      q_vec_subtract( diff, sens_srod_POS, sens_srod_ORI );

#if 0  /* POSITION CHECK */
	      printf( "position methods differ by %lf\n",
		      q_vec_magnitude( diff ) ) ;
#endif

#if 0
	      q_xyz_quat_invert( &sensFsrod, &srodFsens );
	      q_xyz_quat_compose( &asceFsrod_ORI, &asceFsens[ptCount],
				  &sensFsrod );
#else
	      q_invert( sensFsrod.quat, srodFsens.quat );
	      q_mult( asceFsrod_ORI.quat, asceFsens[ptCount].quat,
		      sensFsrod.quat );
#endif
	      q_vec_subtract( diff, asceFsrod_ORI.xyz, asceFsrod.xyz );
	      qx_diff( diffQuat, asceFsrod_ORI.quat, asceFsrod.quat );
	      q_normalize( diffQuat, diffQuat );
	      printf( "angle diff = %lf\n",
		      Q_RAD_TO_DEG( 2.0 * acos( diffQuat[ Q_W ] ) ) ) ;

	    }
#endif
	    /* Replace the quaternion read in with the one we want to plot */
	    q_normalize( asceFsens[ ptCount ].quat, srodFsens.quat );

	    /* Replace the position read with the one we want in the table */
#ifdef ABS_POSN
	    q_vec_copy( faroFrrodu[ ptCount ].xyz, framFsrod.xyz );
#else
	    q_vec_copy( faroFrrodu[ ptCount ].xyz, asceFsrod.xyz );
#endif
	    if( cR->verbose ) {
	      fprintf( cR->outfp, "TRUE XYZ  = ( %lf %lf %lf )\n",
		       faroFrrodu[ ptCount ].xyz[Q_X],
		       faroFrrodu[ ptCount ].xyz[Q_Y],
                       faroFrrodu[ ptCount ].xyz[Q_Z] );
	      q_normalize( asceFsrod.quat, asceFsrod.quat );
	      fprintf( cR->outfp, "TRUE QUAT = ( %lf %lf %lf %lf )\n",
		       asceFsrod.quat[Q_X], asceFsrod.quat[Q_Y],
		       asceFsrod.quat[Q_Z], asceFsrod.quat[Q_W] );
	    }

	    /****************************************************************
	    ** At this point, the tuple we want is (10 numbers: 3, 3, 4)
	    **  { asceFsens[ ptCount ].xyz,       Position read in ASCE
            **    faroFrrodu[ ptCount ].xyz,      Correct position in ASCE
            **    asceFsens[ ptCount ].quat }     Orientation correction
	    ****************************************************************/

	    /* Extend the bounding box */
	    if( asceFsens[ ptCount ].xyz[Q_X] < minSamp[Q_X] )
	      minSamp[Q_X] = asceFsens[ ptCount ].xyz[Q_X];
	    else if( asceFsens[ ptCount ].xyz[Q_X] > maxSamp[Q_X] )
	      maxSamp[Q_X] = asceFsens[ ptCount ].xyz[Q_X];
	    if( asceFsens[ ptCount ].xyz[Q_Y] < minSamp[Q_Y] )
	      minSamp[Q_Y] = asceFsens[ ptCount ].xyz[Q_Y];
	    else if( asceFsens[ ptCount ].xyz[Q_Y] > maxSamp[Q_Y] )
	      maxSamp[Q_Y] = asceFsens[ ptCount ].xyz[Q_Y];
	    if( asceFsens[ ptCount ].xyz[Q_Z] < minSamp[Q_Z] )
	      minSamp[Q_Z] = asceFsens[ ptCount ].xyz[Q_Z];
	    else if( asceFsens[ ptCount ].xyz[Q_Z] > maxSamp[Q_Z] )
	      maxSamp[Q_Z] = asceFsens[ ptCount ].xyz[Q_Z];

	    /* Count the point */
	    ptCount++;
	  } /* end if( point is OK ) */

	} /* end for( ptsLeft ) */

	/* Valid points are asceFsens[ ] and faroFrrodu[ptLastCt..ptCount] */
	if( cR->verbose )
	  printf( "Accepted %d out of %d points.\n",
	          ptCount - ptLastCt, cR->nPts[j] );
	ptLastCt = ptCount;

#ifdef ASCII_FILES
	if( CloseFile( ) ) 
#else
	if( CloseBFile( ) ) {
#endif
	  fprintf( stderr, "Error closing file '%s'\n", cR->sampFile[j] );
	} else if( cR->verbose )
	  printf( "File closed.\n" );
       }
    } /* end for( each file ) */

    fclose( fobPOSnoiseVdist );
    fclose( fobORInoiseVdist );
    fclose( faroPOSnoiseVdist );
    fclose( faroORInoiseVdist );

    /* A few statistics */
    q_vec_add( ctrSamp, minSamp, maxSamp );
    q_vec_scale( ctrSamp, 0.5, ctrSamp );
    printf( "%d sample points accepted (%d by aux counter)\n",
	    ptCount, rejPts[0] );
    printf( "Sample minima: " );    q_vec_print( minSamp );
    printf( "Sample maxima: " );    q_vec_print( maxSamp );
    printf( "Sample centroid: " );  q_vec_print( ctrSamp );
    radSamp = 0.5 * q_vec_distance( minSamp, maxSamp );
    printf( "Sample radius: %8lf\n", radSamp );
    /* rejected for one reason */
    printf( "%d points rejected for FOB position jitter\n",
	    rejPts[ fobPOSN_JITTER ] );
    printf( "%d points rejected for FOB orientation jitter\n",
	    rejPts[ fobANG_JITTER ] );
    printf( "%d points rejected for Far position jitter\n",
	    rejPts[ farPOSN_JITTER ] );
    printf( "%d points rejected for Far orientation jitter\n",
	    rejPts[ farANG_JITTER ] );
    /* rejected for two reasons */
    printf( "%d points rejected for FOB position and orientation jitter\n",
	    rejPts[ fobPOSN_JITTER + fobANG_JITTER ] );
    printf( "%d points rejected for Faro position and orientation jitter\n",
	    rejPts[ farPOSN_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB/Faro position jitter\n",
	    rejPts[ fobPOSN_JITTER + farPOSN_JITTER ] );
    printf( "%d points rejected for FOB/Faro orientation jitter\n",
	    rejPts[ fobANG_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB pos / Faro ori jitter\n",
	    rejPts[ fobPOSN_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB ori / Faro pos jitter\n",
	    rejPts[ fobANG_JITTER + farPOSN_JITTER ] );
    /* rejected for three reasons */
    printf( "%d points rejected for FOB ori / Far pos-ori jitter\n",
	    rejPts[ fobANG_JITTER + farPOSN_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB pos / Far pos-ori jitter\n",
	    rejPts[ fobPOSN_JITTER + farPOSN_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB pos-ori / Far ori jitter\n",
	    rejPts[ fobPOSN_JITTER + fobANG_JITTER + farANG_JITTER ] );
    printf( "%d points rejected for FOB pos-ori / Far pos jitter\n",
	    rejPts[ fobPOSN_JITTER + fobANG_JITTER + farPOSN_JITTER ] );
    /* rejected for four reasons */
    printf(
    "%d points rejected for FOB/Far position and FOB/Far orientation jitter\n",
	    rejPts[ fobPOSN_JITTER + fobANG_JITTER +
		    farPOSN_JITTER + farANG_JITTER ] );

    return ptCount;
}
